"""Tests for the Topic class."""

from typing import Any, List

import pytest

from kafkaconnect.topic_names_set import TopicNamesSet

Fixture = Any


@pytest.fixture()
def topic_names_list() -> List:
    """Return a list of topic names to use in the tests."""
    return ["test.t1", "test.t2", "test.t3"]


def test_select_regex(topic_names_list: Fixture) -> None:
    """Test topic selection using a regex."""
    t = TopicNamesSet(topic_names_list, select_regex="test.*")

    assert "test.t1" in t.topic_names_set
    assert "test.t2" in t.topic_names_set
    assert "test.t3" in t.topic_names_set


def test_select_regex_topic_name(topic_names_list: Fixture) -> None:
    """Test topic selection using a topic name."""
    t = TopicNamesSet(topic_names_list, select_regex="test.t1")

    assert "test.t1" in t.topic_names_set
    assert "test.t2" not in t.topic_names_set
    assert "test.t3" not in t.topic_names_set


def test_exclude_regex(topic_names_list: Fixture) -> None:
    """Test topic exlusion using a regex."""
    t = TopicNamesSet(
        topic_names_list, select_regex="test.*", exclude_regex="test.t[1-2]"
    )

    assert "test.t1" not in t.topic_names_set
    assert "test.t2" not in t.topic_names_set
    assert "test.t3" in t.topic_names_set


def test_exclude_regex_topic_name(topic_names_list: Fixture) -> None:
    """Test topic exclusion using a single topic name."""
    t = TopicNamesSet(
        topic_names_list, select_regex="test.*", exclude_regex="test.t2"
    )

    assert "test.t1" in t.topic_names_set
    assert "test.t2" not in t.topic_names_set
    assert "test.t3" in t.topic_names_set


def test_exclude_all(topic_names_list: Fixture) -> None:
    """Test selecting and excluding all topics."""
    t = TopicNamesSet(
        topic_names_list, select_regex="test.*", exclude_regex="test.*"
    )

    assert t.topic_names_set == set()



"""Tets for the connect module."""

import pytest

from kafkaconnect.connect import Connect, HTTPMethod
from kafkaconnect.influxdb_sink.config import InfluxConfig


def test_expected_exception_get() -> None:
    """Test expected exception."""
    with pytest.raises(ValueError):
        connect = Connect(connect_url="http://some-url")
        connect._request(
            HTTPMethod.GET, uri="http://some-url", data='{"foo": "bar"}'
        )


def test_expected_exception_delete() -> None:
    """Test expected exception."""
    with pytest.raises(ValueError):
        connect = Connect(connect_url="http://some-url")
        connect._request(
            HTTPMethod.DELETE, uri="http://some-url", data='{"foo": "bar"}'
        )


@pytest.mark.vcr
def test_list() -> None:
    """Test list method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.list()
    assert "influxdb-sink" in result


@pytest.mark.vcr
def test_info() -> None:
    """Test info method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.info("influxdb-sink")
    assert '"name": "influxdb-sink"' in result


@pytest.mark.vcr
def test_status() -> None:
    """Test status method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.status("influxdb-sink")
    assert "RUNNING" in result


@pytest.mark.vcr
def test_config() -> None:
    """Test config method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.config("influxdb-sink")
    assert '"name": "influxdb-sink"' in result


@pytest.mark.vcr
def test_tasks() -> None:
    """Test tasks method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.tasks("influxdb-sink")
    assert '"task": 0' in result


@pytest.mark.vcr
def test_topics() -> None:
    """Test topics method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.topics("influxdb-sink")
    # Kafka Connect 5.3.1
    assert "connectors/influxdb-sink/topics not found." in result


@pytest.mark.vcr
def test_plugins() -> None:
    """Test plugins method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.plugins()
    assert "InfluxSinkConnector" in result


@pytest.mark.vcr
def test_remove() -> None:
    """Test remove method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.remove("influxdb-sink")
    assert result == ""


@pytest.mark.vcr
def test_create_or_update() -> None:
    """Test create_or_update method."""
    connect = Connect(connect_url="http://localhost:8083")
    connect_config = InfluxConfig(
        name="influxdb-sink",
        connect_influx_url="http://localhost:8086",
        connect_influx_db="mydb",
        tasks_max=1,
        connect_influx_username="foo",
        connect_influx_password="bar",
        connect_influx_error_policy="foo",
        connect_influx_max_retries="1",
        connect_influx_retry_interval="1",
        connect_progress_enabled=True,
        tags="",
        remove_prefix="",
    )
    connect_config.update_config({"t1", "t2", "t3"})
    result = connect.create_or_update(
        name="influxdb-sink", connect_config=connect_config.asjson()
    )
    assert "influxdb-sink" in result


@pytest.mark.vcr
def test_validate() -> None:
    """Test validate method."""
    connect = Connect(connect_url="http://localhost:8083")
    connect_config = InfluxConfig(
        name="influxdb-sink",
        connect_influx_url="http://localhost:8086",
        connect_influx_db="mydb",
        tasks_max=1,
        connect_influx_username="foo",
        connect_influx_password="bar",
        connect_influx_error_policy="foo",
        connect_influx_max_retries="1",
        connect_influx_retry_interval="1",
        connect_progress_enabled=True,
        tags="",
        remove_prefix="",
    )
    connect_config.update_config({"t1", "t2", "t3"})
    result = connect.validate(
        name="InfluxSinkConnector", connect_config=connect_config.asjson()
    )
    assert "error_count" in result


@pytest.mark.vcr
def test_restart() -> None:
    """Test restart method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.restart("influxdb-sink")
    assert result == ""


@pytest.mark.vcr
def test_pause() -> None:
    """Test pause method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.pause("influxdb-sink")
    assert result == ""


@pytest.mark.vcr
def test_resume() -> None:
    """Test resume method."""
    connect = Connect(connect_url="http://localhost:8083")
    result = connect.resume("influxdb-sink")
    assert result == ""



"""Helper class for interacting with the `Kafka Connect REST Interface.

See https://docs.confluent.io/current/connect/references/restapi.html.
"""

__all__ = ["Connect"]

import json
from enum import Enum
from typing import Optional

from requests import delete, get, post, put  # noqa
from requests.exceptions import ConnectionError, HTTPError


class HTTPMethod(Enum):
    """HTTP methods allowed."""

    GET = "get"
    PUT = "put"
    POST = "post"
    DELETE = "delete"


class Connect:
    """Kafka Connect API helper class.

    Parameters
    ----------
    connect_url : `str`
    Kafka Connect URL
    """

    _header = {"Content-Type": "application/json"}

    def __init__(self, connect_url: str) -> None:
        self._connect_url = connect_url

    def _request(
        self, method: HTTPMethod, uri: str, data: Optional[str] = None
    ) -> str:
        """Make HTTP requests.

        Parameters
        ----------
        method: `HTTPMethod`
            HTTP method as defined in the HTTPMethod class.
        uri : `str`
            The resource identifier.
        data : `str`
            The message body for the PUT request.

        Returns
        -------
        content: `ContenT` or `None`
            The response content. Returns `None` if the request was not
            successful or if the response is empty.
        """
        if method.name in ("GET", "DELETE"):
            if data:
                raise ValueError(
                    f"data argument must be None with {method.name} method."
                )
        func = eval(method.value)
        try:
            if data:
                response = func(uri, data=data, headers=Connect._header)
            else:
                response = func(uri)
            response.raise_for_status()
        except HTTPError as err:
            if err.response.status_code == 404:
                message = f"Resource {uri} not found."
                return message
            # returns 409 (Conflict) if kafka cluster rebalance is in process.
            if err.response.status_code == 409:
                message = "Kafka cluster rebalance is in process."
                return message
        except ConnectionError:
            message = (
                f"Failed to establish connection with the "
                f"Connect API {self._connect_url}."
            )
            return message
        content = ""
        if response.text:
            content = json.dumps(response.json(), indent=4, sort_keys=True)
        return content

    def list(self) -> str:
        """Get a list of active connectors."""
        uri = f"{self._connect_url}/connectors"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def info(self, name: str) -> str:
        """Get information about the connector."""
        uri = f"{self._connect_url}/connectors/{name}"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def status(self, name: str) -> str:
        """Get the connector status."""
        uri = f"{self._connect_url}/connectors/{name}/status"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def config(self, name: str) -> str:
        """Get the connector configuration."""
        uri = f"{self._connect_url}/connectors/{name}/config"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def tasks(self, name: str) -> str:
        """Get a list of tasks currently running for the connector."""
        uri = f"{self._connect_url}/connectors/{name}/tasks"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def topics(self, name: str) -> str:
        """Get the list of topic names used by the connector."""
        uri = f"{self._connect_url}/connectors/{name}/topics"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def plugins(self) -> str:
        """Get a list of connector plugins available in the Connect cluster."""
        uri = f"{self._connect_url}/connector-plugins"
        return self._request(method=HTTPMethod.GET, uri=uri)

    def create_or_update(self, name: str, connect_config: str) -> str:
        """Create or update a connector.

        Create a new connector using the given configuration, or update the
        configuration for an existing connector.

        Parameters
        ----------
        name : `str`
            Connector name.
        connect_config : `str`
            Connector configuration.

        Returns
        -------
        content: `ContentT` or `None`
            The response content. Returns `None` if the request was not
        successful.
        """
        uri = f"{self._connect_url}/connectors/{name}/config"
        return self._request(
            method=HTTPMethod.PUT, uri=uri, data=connect_config
        )

    def restart(self, name: str) -> str:
        """Restart the connector."""
        uri = f"{self._connect_url}/connectors/{name}/restart"
        return self._request(method=HTTPMethod.POST, uri=uri)

    def pause(self, name: str) -> str:
        """Pause the connector."""
        uri = f"{self._connect_url}/connectors/{name}/pause"
        return self._request(method=HTTPMethod.PUT, uri=uri)

    def resume(self, name: str) -> str:
        """Resume a paused connector."""
        uri = f"{self._connect_url}/connectors/{name}/resume"
        return self._request(method=HTTPMethod.PUT, uri=uri)

    def validate(self, name: str, connect_config: str) -> str:
        """Validate the connector configuration.

        Validate the configuration values against the configuration definition.
        """
        uri = f"{self._connect_url}/connector-plugins/{name}/config/validate"
        return self._request(
            method=HTTPMethod.PUT, uri=uri, data=connect_config
        )

    def remove(self, name: str) -> str:
        """Delete a connector, halting tasks and deleting its configuration."""
        uri = f"{self._connect_url}/connectors/{name}"
        return self._request(method=HTTPMethod.DELETE, uri=uri)

    def validate_and_create(self, name: str, connect_config: str) -> str:
        """Validate the configuration before creating the connector."""
        validation = self.validate(
            name=json.loads(connect_config)["connector.class"],
            connect_config=connect_config,
        )

        error_count = json.loads(validation)["error_count"]
        if error_count > 0:
            return validation

        return self.create_or_update(name, connect_config)



"""CLI to create the S3 Sink connector.

See https://docs.confluent.io/current/connect/kafka-connect-s3
"""

__all__ = ["create_s3_sink"]

import json
import time

import click

from kafkaconnect.connect import Connect


@click.command("s3-sink")
@click.argument("configfile")
@click.option(
    "-n",
    "--name",
    "name",
    default=None,
    help=(
        "Name of the S3 Sink connector. "
        "If provided it overrrides the ``name`` property "
        "value in the configuration file."
    ),
)
@click.option(
    "--aws-access-key-id",
    "aws_access_key_id",
    envvar="AWS_ACCESS_KEY_ID",
    default=None,
    help=(
        "The AWS access key ID used to authenticate personal AWS credentials."
    ),
)
@click.option(
    "--aws-secret-access-key",
    "aws_secret_access_key",
    envvar="AWS_SECRET_ACCESS_KEY",
    default=None,
    help=(
        "The secret access key used to authenticate personal AWS credentials."
    ),
)
@click.option(
    "--dry-run",
    "dry_run",
    is_flag=True,
    help=("Validates the connector configuration without creating."),
)
@click.option(
    "--show-status",
    "show_status",
    is_flag=True,
    help=(
        "Show connector status in the output. See also the "
        "``--show-status-interval`` option."
    ),
)
@click.option(
    "--show-status-interval",
    "show_status_interval",
    default=15000,
    show_default=True,
    help=("The time interval in milliseconds to output the connector status."),
)
@click.pass_context
def create_s3_sink(
    ctx: click.Context,
    configfile: str,
    name: str,
    aws_access_key_id: str,
    aws_secret_access_key: str,
    dry_run: bool,
    show_status: bool,
    show_status_interval: int,
) -> int:
    """Create an instance of the S3 Sink connector.

    Use the --show-status option to output status.
    """
    # Get configuration from the parent command
    if ctx.parent:
        parent_config = ctx.parent.obj["config"]

    connect = Connect(parent_config.connect_url)

    with open(configfile) as f:
        config = json.load(f)

    if name:
        click.echo("Updating connector name.")
        config["name"] = name

    if None in (aws_access_key_id, aws_secret_access_key):
        click.echo(
            "Could not get the AWS credentials. "
            "Use the --access-key-id and --aws-secret-access-key options "
            "or set the AWS credentials using the AWS_ACCESS_KEY_ID and "
            "AWS_SECRET_ACCESS_KEY env variables."
        )
        return 1

    config["aws.access.key.id"] = aws_access_key_id
    config["aws.secret.access.key"] = aws_secret_access_key

    # Validate the configuration only.
    if dry_run:
        validation = connect.validate(
            name=config["connector.class"],
            connect_config=json.dumps(config),
        )
        click.echo(validation)
        return 0

    name = config["name"]
    click.echo(f"Creating the {name} connector...")
    click.echo(connect.validate_and_create(name, json.dumps(config)))
    if show_status:
        while True:
            time.sleep(int(show_status_interval) / 1000)
            try:
                click.echo(connect.status(name=name))
            except KeyboardInterrupt:
                raise click.ClickException("Interruped.")

    return 0



"""Kafkaconnect Command Line Interface."""

__all__ = (
    "main",
    "list",
    "info",
    "status",
    "config",
    "tasks",
    "topics",
    "plugins",
    "pause",
    "resume",
    "delete",
    "help",
    "create",
)

import json
from typing import Any, Optional

import click

from kafkaconnect.config import Config
from kafkaconnect.connect import Connect
from kafkaconnect.influxdb_sink.cli import create_influxdb_sink
from kafkaconnect.jdbc_sink.cli import create_jdbc_sink
from kafkaconnect.mirrormaker2.cli import create_mirrormaker2
from kafkaconnect.s3_sink.cli import create_s3_sink

# Add -h as a help shortcut option
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-b",
    "--broker",
    "broker_url",
    envvar="KAFKA_BROKER_URL",
    default="localhost:9092",
    show_default=True,
    help="Kafka Broker URL. Alternatively set via $KAFKA_BROKER_URL env var.",
)
@click.option(
    "-c",
    "--connect",
    "connect_url",
    envvar="KAFKA_CONNECT_URL",
    default="http://localhost:8083",
    show_default=True,
    help=(
        "Kafka Connect URL. Alternatively set via $KAFKA_CONNECT_URL env var."
    ),
)
@click.option(
    "-u",
    "--username",
    "sasl_plain_username",
    envvar="KAFKA_USERNAME",
    default=None,
    show_default=True,
    help=("Username for SASL authentication."),
)
@click.option(
    "-p",
    "--password",
    "sasl_plain_password",
    envvar="KAFKA_PASSWORD",
    default=None,
    show_default=True,
    help=("Password for SASL authentication."),
)
@click.version_option(message="%(version)s")
@click.pass_context
def main(
    ctx: click.Context,
    broker_url: str,
    connect_url: str,
    sasl_plain_username: str,
    sasl_plain_password: str,
) -> None:
    """Command-line interface for kafkaconnect.

    kafkaconnect is a Connect API client that helps to configure and
    manage Kafka connectors.
    """
    config = Config(
        broker_url=broker_url,
        connect_url=connect_url,
        sasl_plain_username=sasl_plain_username,
        sasl_plain_password=sasl_plain_password,
    )
    ctx.ensure_object(dict)
    ctx.obj["config"] = config


@main.command("list")
@click.pass_context
def list(ctx: click.Context) -> None:
    """Get a list of active connectors."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.list())


@main.command("info")
@click.argument("name")
@click.pass_context
def info(ctx: click.Context, name: str) -> None:
    """Get information about the connector."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.info(name))


@main.command("status")
@click.argument("name")
@click.pass_context
def status(ctx: click.Context, name: str) -> None:
    """Get the connector status."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.status(name))


@main.command("config")
@click.argument("name")
@click.pass_context
def config(ctx: click.Context, name: str) -> None:
    """Get the connector configuration."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.config(name))


@main.command("tasks")
@click.argument("name")
@click.pass_context
def tasks(ctx: click.Context, name: str) -> None:
    """Get a list of tasks currently running for the connector."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.tasks(name))


@main.command("topics")
@click.argument("name")
@click.pass_context
def topics(ctx: click.Context, name: str) -> None:
    """Get the list of topic names used by the connector."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.topics(name))


@main.command("plugins")
@click.pass_context
def plugins(ctx: click.Context) -> None:
    """Get a list of connector plugins available in the Connect cluster."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.plugins())


@main.command("restart")
@click.argument("name")
@click.pass_context
def restart(ctx: click.Context, name: str) -> None:
    """Restart a connector and its tasks."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.restart(name))


@main.command("pause")
@click.argument("name")
@click.pass_context
def pause(ctx: click.Context, name: str) -> None:
    """Pause the connector and its tasks."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.pause(name))


@main.command("resume")
@click.argument("name")
@click.pass_context
def resume(ctx: click.Context, name: str) -> None:
    """Resume a paused connector."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.resume(name))


@main.command("delete")
@click.argument("name")
@click.pass_context
def delete(ctx: click.Context, name: str) -> None:
    """Delete a connector.

    Halt tasks and remove the connector configuration.
    """
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)
    click.echo(connect.remove(name))


@main.command("upload")
@click.argument("configfile")
@click.option(
    "-n",
    "--name",
    "name",
    required=True,
    help=("Name of the connector."),
)
@click.option(
    "--dry-run",
    is_flag=True,
    help=("Validate the connector configuration without uploading."),
)
@click.pass_context
def upload(
    ctx: click.Context,
    configfile: str,
    name: str,
    dry_run: bool,
) -> int:
    """Upload the connector configuration from a file."""
    config = ctx.obj["config"]
    connect = Connect(config.connect_url)

    with open(configfile) as f:
        connect_config = json.load(f)

    # Ensure connector name is consistent
    connect_config["name"] = name

    # Validate the connector configuration only.
    if dry_run:
        validation = connect.validate(
            name=connect_config["connector.class"],
            connect_config=json.dumps(connect_config),
        )
        click.echo(validation)
        return 0

    click.echo(f"Uploading {name} connector configuration...")
    click.echo(connect.validate_and_create(name, json.dumps(connect_config)))
    return 0


@main.command()
@click.argument("topic", default=None, required=False, nargs=1)
@click.pass_context
def help(ctx: click.Context, topic: Optional[str], **kw: Any) -> None:
    """Show help for any command."""
    # The help command implementation is taken from
    # https://www.burgundywall.com/post/having-click-help-subcommand
    if topic:
        if topic in main.commands:
            ctx.info_name = topic
            click.echo(main.commands[topic].get_help(ctx))
        else:
            raise click.UsageError(f"Unknown help topic {topic}", ctx)
    else:
        assert ctx.parent
        click.echo(ctx.parent.get_help())


@main.group()
@click.pass_context
def create(ctx: click.Context) -> None:
    """Create a new connector.

    Each subcommand creates a different connector.
    """


# Add subcommands from other modules
create.add_command(create_influxdb_sink)
create.add_command(create_s3_sink)
create.add_command(create_mirrormaker2)
create.add_command(create_jdbc_sink)



"""A set of topics names used to configure the connector."""

__all__ = ["TopicNamesSet"]

import logging
import re
from typing import List, Optional, Set, Type, TypeVar

from confluent_kafka import KafkaException
from confluent_kafka.admin import AdminClient

from kafkaconnect.config import Config

T = TypeVar("T", bound="TopicNamesSet")


logger = logging.getLogger("kafkaconnect")

SECURITY_PROTOCOL = "SASL_PLAINTEXT"
SASL_MECHANISM = "SCRAM-SHA-512"


class TopicNamesSet:
    """A set of topics names used to configure the connector.

    Parameters
    ----------
    topic_names_list : `List`
        A list topic names
    select_regex : `str`
        A regex to add topic names to the set.
    exclude_regex : `str`
        A regex to exclude topic names from the set.


    Raises
    ------
    ValueError
        If an ``topic_names_list`` entry contains any invalid names
        or duplicates with prior information.

    """

    def __init__(
        self,
        topic_names_list: List,
        select_regex: str = ".*",
        exclude_regex: Optional[str] = None,
    ) -> None:
        self.topic_names_list = topic_names_list
        self.select_regex = select_regex
        self.exclude_regex = exclude_regex
        self.topic_names_set = self.filter_topics()

    def filter_topics(self) -> Set[str]:
        """Filter a list of topic names.

        Returns
        -------
        topic_names_set: `Set`
            A a set of topic names.
        """
        topic_names_set = set(self.topic_names_list)

        if self.select_regex:
            pattern = re.compile(self.select_regex)
            topic_names_set = {
                topic for topic in topic_names_set if pattern.match(topic)
            }

        if self.exclude_regex:
            pattern = re.compile(self.exclude_regex)
            excluded_topic_names = {
                topic for topic in topic_names_set if pattern.match(topic)
            }
            topic_names_set = topic_names_set - excluded_topic_names

        return topic_names_set

    @classmethod
    def from_kafka(
        cls: Type[T],
        config: Config,
        select_regex: str = ".*",
        exclude_regex: Optional[str] = None,
    ) -> T:
        """Create the topic name set from a list of topic names in Kafka."""
        if config.sasl_plain_username and config.sasl_plain_password:
            kafka_config = {
                "bootstrap.servers": config.broker_url,
                "security.protocol": SECURITY_PROTOCOL,
                "sasl.mechanisms": SASL_MECHANISM,
                "sasl.username": config.sasl_plain_username,
                "sasl.password": config.sasl_plain_password,
            }
        elif (
            config.sasl_plain_username is None
            and config.sasl_plain_password is None
        ):
            kafka_config = {"bootstrap.servers": config.broker_url}
        else:
            raise ValueError(
                "Both or neither of 'config.sasl_plain_username' "
                "and 'config.sasl_plain_password' must be set."
            )
        try:
            broker_client = AdminClient(kafka_config)
        except KafkaException:
            message = (
                f"Failed to establish connection with broker "
                f"{config.broker_url}."
            )
            logger.error(message)

        topic_names_list: List = broker_client.list_topics(
            timeout=10
        ).topics.keys()

        return cls(
            topic_names_list=topic_names_list,
            select_regex=select_regex,
            exclude_regex=exclude_regex,
        )



"""InfluxDB Sink connector configuration.

See https://docs.lenses.io/connectors/sink/influx.html.
"""

__all__ = ["InfluxConfig"]

from dataclasses import dataclass
from typing import Set

from kafkaconnect.config import ConnectorConfig


@dataclass
class InfluxConfig(ConnectorConfig):
    """InfluxDB connector configuration."""

    name: str
    """Name of the connector."""

    connect_influx_url: str
    """InfluxDB connection URL."""

    connect_influx_db: str
    """InfluxDB database name."""

    tasks_max: int
    """Number of Kafka Connect tasks."""

    connect_influx_username: str
    """InfluxDB username."""

    connect_influx_password: str
    """InfluxDB password."""

    connect_influx_error_policy: str
    """Connector error policy configuration."""

    connect_influx_max_retries: str
    """Connector error policy configuration."""

    connect_influx_retry_interval: str
    """Connector error policy configuration."""

    connect_progress_enabled: bool
    """Enables the output for how many records have been processed."""

    tags: str
    """Fields to be used as tags."""

    remove_prefix: str
    """Prefix to remove from topic name to use as measurement name."""

    # Attributes with defaults are not configurable via click
    topics: str = ""
    """Comma separated list of Kafka topics to read from."""

    connect_influx_kcql: str = ""
    """KCQL queries to extract fields from topics. Computed.

    We assume that a topic has a flat structure so that `SELECT * FROM` will
    retrieve all topic fields. This is configuration is derived from the list
    of topics and from the timestamp to use as the InfluxDB time.
    """

    connector_class: str = (
        "com.datamountaineer.streamreactor.connect.influx.InfluxSinkConnector"
    )
    """Stream reactor InfluxDB Sink connector class."""

    def update_config(self, topics: Set[str], timestamp: str = "") -> None:
        """Update connector config.

        Parameters
        ----------
        topics : `Set`
            List of kafka topics.

        timestamp : `str`
            Timestamp used as influxDB time.
        """
        sorted_topics = sorted(topics)
        self.topics = ",".join(sorted_topics)

        tags = ""
        if self.tags:
            tags = f" WITHTAG({self.tags})"

        queries = []
        for topic in sorted_topics:
            if self.remove_prefix:
                measurement = topic[len(self.remove_prefix) :]
            else:
                measurement = topic

            query = (
                f"INSERT INTO {measurement} SELECT * FROM {topic} "
                f"WITHTIMESTAMP {timestamp} TIMESTAMPUNIT=MICROSECONDS{tags}"
            )

            queries.append(query)

        self.connect_influx_kcql = ";".join(queries)



"""CLI to create the MirrorMaker 2 connectors.

https://cwiki.apache.org/confluence/display/KAFKA/KIP-382%3A+MirrorMaker+2.0
"""

__all__ = ["create_mirrormaker2"]

import json
import time

import click

from kafkaconnect.connect import Connect


@click.command("mirrormaker2")
@click.option(
    "-n",
    "--name",
    "name",
    default=None,
    show_default=True,
    help=(
        "Name of the MirroMaker 2 instance. "
        "If provided it is used as a prefix to name the heartbeat, checkpoint "
        "and mirror-source connectors and it overrides the ``name`` property "
        "value in the configuration file."
    ),
)
@click.option(
    "-h",
    "--heartbeat",
    "heartbeat_configfile",
    required=True,
    help=("Hertbeat connector configuration file."),
)
@click.option(
    "-c",
    "--checkpoint",
    "checkpoint_configfile",
    required=True,
    help=("Checkpoint connector configuration file."),
)
@click.option(
    "-m",
    "--mirror-source",
    "mirror_source_configfile",
    required=True,
    help=("MirrorSource connector configuration file."),
)
@click.option(
    "--dry-run",
    "dry_run",
    is_flag=True,
    help=("Validates the connector configuration without creating."),
)
@click.option(
    "--show-status",
    "show_status",
    is_flag=True,
    help=(
        "Show connector status in the output. See also the "
        "``--show-status-interval`` option."
    ),
)
@click.option(
    "--show-status-interval",
    "show_status_interval",
    default=15000,
    show_default=True,
    help=("The time interval in milliseconds to output the connector status."),
)
@click.pass_context
def create_mirrormaker2(
    ctx: click.Context,
    name: str,
    heartbeat_configfile: str,
    checkpoint_configfile: str,
    mirror_source_configfile: str,
    dry_run: bool,
    show_status: bool,
    show_status_interval: int,
) -> int:
    """Create an instance of the MirrorMaker 2 connectors.

    Create the heartbeat, checkpoint and mirror-source
    connectors. Use the --show-status option to output status.
    """
    # Get configuration from the main command
    if ctx.parent:
        config = ctx.parent.obj["config"]

    connect = Connect(config.connect_url)

    with open(heartbeat_configfile) as f:
        heartbeat_config = json.load(f)

    with open(checkpoint_configfile) as f:
        checkpoint_config = json.load(f)

    with open(mirror_source_configfile) as f:
        mirror_source_config = json.load(f)

    # Override connector name in the configuration
    if name:
        heartbeat_config["name"] = f"{name}-heartbeat"
        checkpoint_config["name"] = f"{name}-checkpoint"
        mirror_source_config["name"] = f"{name}-mirror-source"

    # Validate the configuration only.
    if dry_run:
        heartbeat_validation = connect.validate(
            name=heartbeat_config["connector.class"],
            connect_config=json.dumps(heartbeat_config),
        )
        click.echo(heartbeat_validation)
        checkpoint_validation = connect.validate(
            name=checkpoint_config["connector.class"],
            connect_config=json.dumps(checkpoint_config),
        )
        click.echo(checkpoint_validation)
        mirror_source_validation = connect.validate(
            name=mirror_source_config["connector.class"],
            connect_config=json.dumps(mirror_source_config),
        )
        click.echo(mirror_source_validation)
        return 0

    heartbeat_name = heartbeat_config["name"]
    click.echo(f"Creating the {heartbeat_name} connector...")
    click.echo(
        connect.validate_and_create(
            heartbeat_name, json.dumps(heartbeat_config)
        )
    )
    checkpoint_name = checkpoint_config["name"]
    click.echo(f"Creating the {checkpoint_name} connector...")
    click.echo(
        connect.validate_and_create(
            checkpoint_name, json.dumps(checkpoint_config)
        )
    )
    mirror_source_name = mirror_source_config["name"]
    click.echo(f"Creating the {mirror_source_name} connector...")
    click.echo(
        connect.validate_and_create(
            mirror_source_name, json.dumps(mirror_source_config)
        )
    )

    if show_status:
        while True:
            time.sleep(int(show_status_interval) / 1000)
            try:
                click.echo(connect.status(name=heartbeat_name))
                click.echo(connect.status(name=checkpoint_name))
                click.echo(connect.status(name=mirror_source_name))
            except KeyboardInterrupt:
                raise click.ClickException("Interruped.")

    return 0



"""Kafkaconnect and connector configuration."""

__all__ = ["Config", "ConnectorConfig"]

import json
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Set, Tuple


@dataclass
class Config:
    """Application configuration."""

    broker_url: str
    """The Kafka Broker URL.
    """

    connect_url: str
    """The Kafka Connect URL.

    The Kafka Connect REST API is used to manage connectors.
    """

    sasl_plain_username: Optional[str] = None
    """Username for SASL authentication.
       If specified then you must also specify sasl_plain_password.
       Default: None
    """

    sasl_plain_password: Optional[str] = None
    """Password for SASL authentication.
       If specified then you must also specify sasl_plain_username.
       Default: None
    """

    def __post_init__(self) -> None:
        """Post init validation."""
        if (self.sasl_plain_username is None) != (
            self.sasl_plain_password is None
        ):
            raise ValueError(
                "Both or neither of 'sasl_plain_username' "
                "and 'sasl_plain_password' must be specified."
            )


@dataclass
class ConnectorConfig(ABC):
    """Connector configuration interface."""

    @abstractmethod
    def update_config(self, topics: Set[str]) -> None:
        """Update connector configuration.

        Abstract method to be implemented by subclasses.
        """
        pass

    @staticmethod
    def format_field_names(fields: List[Tuple[str, Any]]) -> Dict[str, str]:
        """Rename a field name by replacing '_' with '.'.

        Dictionary factory used with the dataclasses.asdict() method.
        """
        result = []
        for f in fields:
            name, value = f
            name = name.replace("_", ".")
            result.append((name, value))
        return dict(result)

    def asjson(self) -> str:
        """Convert dataclass instance into JSON."""
        config = asdict(self, dict_factory=self.format_field_names)
        return json.dumps(config, indent=4, sort_keys=True)



"""Tests for the command-line interface."""

from click.testing import CliRunner

from kafkaconnect.cli import main


def test_help() -> None:
    """Test help for main commands and subcommands."""
    runner = CliRunner()

    result = runner.invoke(main, ["-h"])
    assert result.exit_code == 0
    assert "Commands:" in result.output

    result = runner.invoke(main, ["help"])
    assert result.exit_code == 0
    assert "Commands:" in result.output

    result = runner.invoke(main, ["help", "list"])
    assert result.exit_code == 0
    assert "Commands:" not in result.output
    assert "Options:" in result.output

    result = runner.invoke(main, ["help", "unknown-command"])
    assert result.exit_code != 0
    assert "Unknown help topic unknown-command" in result.output


def test_failed_connection_message() -> None:
    """Fails with a specific message if Connect API URL is invalid.

    Also test passing context with the Connect API URL.
    """
    runner = CliRunner()
    # test failed connection with kafka connect
    result = runner.invoke(
        main, ["--connect", "http://invalid-connect-url", "list"]
    )
    assert result.exit_code == 0
    assert (
        "Failed to establish connection with the Connect "
        "API http://invalid-connect-url.\n" in result.output
    )


def test_create_influxdb_sink() -> None:
    """Test create influxdb-sink connector with default configuration."""
    runner = CliRunner()
    result = runner.invoke(
        main, ["create", "influxdb-sink", "--dry-run", "t1"]
    )
    assert result.exit_code == 0
    # This query is built by InfluxConfig.update_influx_kcql()
    assert (
        '"connect.influx.kcql": '
        '"INSERT INTO t1 SELECT * FROM t1 WITHTIMESTAMP sys_time() '
        'TIMESTAMPUNIT=MICROSECONDS"' in result.output
    )
    # Topics are added by ConnectConfig.update_config()
    assert '"topics": "t1"' in result.output


def test_influxdb_tags() -> None:
    """Test influxdb-sink with tags support."""
    runner = CliRunner()
    result = runner.invoke(
        main, ["create", "influxdb-sink", "--dry-run", "--tags", "test", "t1"]
    )
    assert result.exit_code == 0
    # Add WITHTAG clause to the influx kcql query when using tags
    assert (
        '"connect.influx.kcql": '
        '"INSERT INTO t1 SELECT * FROM t1 WITHTIMESTAMP sys_time() '
        'TIMESTAMPUNIT=MICROSECONDS WITHTAG(test)"' in result.output
    )


def test_remove_prefix() -> None:
    """Test influxdb-sink with remove_prefix support."""
    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "create",
            "influxdb-sink",
            "--dry-run",
            "--remove-prefix",
            "abc.",
            "abc.t1",
        ],
    )
    assert result.exit_code == 0
    # Add WITHTAG clause to the influx kcql query when using tags
    assert (
        '"connect.influx.kcql": '
        '"INSERT INTO t1 SELECT * FROM abc.t1 WITHTIMESTAMP sys_time() '
        'TIMESTAMPUNIT=MICROSECONDS"' in result.output
    )


def test_password_from_env() -> None:
    """Test getting the influxdb password from the environment."""
    env = {"KAFKA_CONNECT_INFLUXDB_PASSWORD": "envpasswd"}
    runner = CliRunner()
    result = runner.invoke(
        main, args=["create", "influxdb-sink", "--dry-run", "t1"], env=env
    )
    assert result.exit_code == 0
    assert "envpasswd" in result.output



"""CLI to create the InfluxDB Sink connector.

See https://docs.lenses.io/connectors/sink/influx.html.
"""

__all__ = ["create_influxdb_sink"]

import json
import time
from typing import Set

import click

from kafkaconnect.connect import Connect
from kafkaconnect.influxdb_sink.config import InfluxConfig
from kafkaconnect.topic_names_set import TopicNamesSet


@click.command("influxdb-sink")
@click.argument("topiclist", nargs=-1, required=False)
@click.option(
    "-n",
    "--name",
    "name",
    envvar="KAFKA_CONNECT_NAME",
    default="influxdb-sink",
    show_default=True,
    help=(
        "Name of the connector."
        "The connector name must be unique accross the cluster."
        "Alternatively set via the $KAFKA_CONNECT_NAME env var."
    ),
)
@click.option(
    "-i",
    "--influxdb_url",
    "connect_influx_url",
    envvar="KAFKA_CONNECT_INFLUXDB_URL",
    default="http://localhost:8086",
    show_default=True,
    help=(
        "InfluxDB connection URL. Alternatively set via the "
        "$KAFKA_CONNECT_INFLUXDB_URL env var."
    ),
)
@click.option(
    "-d",
    "--database",
    "connect_influx_db",
    envvar="KAFKA_CONNECT_DATABASE",
    default="mydb",
    show_default=True,
    help=(
        "InfluxDB database name. The database must exist at InfluxDB. "
        "Alternatively set via the $KAFKA_CONNECT_DATABASE env var."
    ),
)
@click.option(
    "-t",
    "--tasks-max",
    "tasks_max",
    envvar="KAFKA_CONNECT_TASKS_MAX",
    default="1",
    show_default=True,
    help=(
        "Number of Kafka Connect tasks. Alternatively set via the "
        "$KAFKA_CONNECT_TASKS_MAX env var."
    ),
)
@click.option(
    "-u",
    "--username",
    "connect_influx_username",
    envvar="KAFKA_CONNECT_INFLUXDB_USERNAME",
    default="-",
    show_default=True,
    help=(
        "InfluxDB username. Alternatively set via the "
        "$KAFKA_CONNECT_INFLUXDB_USERNAME env var. "
        "Use '-' for unauthenticated users."
    ),
)
@click.option(
    "-p",
    "--password",
    "connect_influx_password",
    envvar="KAFKA_CONNECT_INFLUXDB_PASSWORD",
    default="",
    show_default=True,
    help=(
        "InfluxDB password. Alternatively set via the "
        "$KAFKA_CONNECT_INFLUXDB_PASSWORD env var."
    ),
)
@click.option(
    "-r",
    "--topic-regex",
    "topic_regex",
    envvar="KAFKA_CONNECT_TOPIC_REGEX",
    default=".*",
    show_default=True,
    help=(
        "Regex for selecting topics. Alternatively set via the "
        "$KAFKA_CONNECT_TOPIC_REGEX env var."
    ),
)
@click.option(
    "--dry-run",
    is_flag=True,
    help=(
        "Show the InfluxDB Sink Connector configuration but do not create "
        "the connector."
    ),
)
@click.option(
    "--auto-update",
    is_flag=True,
    help=(
        "Check for new topics and update the connector. See also the "
        "--check-interval option."
    ),
)
@click.option(
    "-v",
    "--validate",
    is_flag=True,
    help="Validate the connector configuration before creating.",
)
@click.option(
    "-c",
    "--check-interval",
    "check_interval",
    envvar="KAFKA_CONNECT_CHECK_INTERVAL",
    default="15000",
    show_default=True,
    help=(
        "The interval, in milliseconds, to check for new topics and update"
        "the connector. Alternatively set via the "
        "$KAFKA_CONNECT_CHECK_INTERVAL env var."
    ),
)
@click.option(
    "-e",
    "--excluded_topic_regex",
    "excluded_topic_regex",
    envvar="KAFKA_CONNECT_EXCLUDED_TOPIC_REGEX",
    default="",
    show_default=True,
    help=(
        "Regex for excluding topics. Alternatively set via the "
        "$KAFKA_CONNECT_EXCLUDED_TOPIC_REGEX env var."
    ),
)
@click.option(
    "--error-policy",
    "connect_influx_error_policy",
    type=click.Choice(["NOOP", "THROW", "RETRY"]),
    envvar="KAFKA_CONNECT_ERROR_POLICY",
    default="THROW",
    show_default=True,
    help=(
        "Specifies the action to be taken if an error occurs while "
        "inserting the data. There are three available options, NOOP, "
        "the error is swallowed, THROW, the error is allowed to propagate "
        "and RETRY. For RETRY the Kafka message is redelivered up to a "
        "maximum number of times specified by the ``--max-retries`` option. "
        "The retry interval is specified by the ``--retry-interval`` option. "
        "Alternatively set via the $KAFKA_CONNECT_ERROR_POLICY env var."
    ),
)
@click.option(
    "--max-retries",
    "connect_influx_max_retries",
    envvar="KAFKA_CONNECT_MAX_RETRIES",
    default="10",
    show_default=True,
    help=(
        "The maximum number of times a message is retried. Only valid when "
        "the ``--error-policy`` is set to RETRY. Alternatively set via the "
        "$KAFKA_CONNECT_MAX_RETRIES env var."
    ),
)
@click.option(
    "--retry-interval",
    "connect_influx_retry_interval",
    envvar="KAFKA_CONNECT_RETRY_INTERVAL",
    default="60000",
    show_default=True,
    help=(
        "The interval, in milliseconds between retries. Only valid when "
        "the ``--error-policy`` is set to RETRY. Alternatively set via the "
        "$KAFKA_CONNECT_RETRY_INTERVAL env var."
    ),
)
@click.option(
    "--progress-enabled",
    "connect_progress_enabled",
    envvar="KAFKA_CONNECT_PROGRESS_ENABLED",
    default="false",
    show_default=True,
    help=(
        "Enables the output for how many records have been processed. "
        "Alternatively set via the $KAFKA_CONNECT_PROGRESS_ENABLED env var."
    ),
)
@click.option(
    "--timestamp",
    "timestamp",
    envvar="KAFKA_CONNECT_INFLUXDB_TIMESTAMP",
    default="sys_time()",
    show_default=True,
    help="Timestamp to use as the InfluxDB time.",
)
@click.option(
    "--tags",
    "tags",
    envvar="KAFKA_CONNECT_INFLUXDB_TAGS",
    default="",
    show_default=True,
    help="Fields in the Avro payload that are treated as InfluxDB tags.",
)
@click.option(
    "--remove-prefix",
    "remove_prefix",
    envvar="KAFKA_CONNECT_INFLUXDB_REMOVE_PREFIX",
    default="",
    show_default=True,
    help="Prefix to remove from topic name to use as measurement name.",
)
@click.pass_context
def create_influxdb_sink(
    ctx: click.Context,
    topiclist: tuple,
    name: str,
    connect_influx_url: str,
    connect_influx_db: str,
    tasks_max: str,
    connect_influx_username: str,
    connect_influx_password: str,
    topic_regex: str,
    dry_run: bool,
    auto_update: bool,
    validate: bool,
    check_interval: str,
    excluded_topic_regex: str,
    connect_influx_error_policy: str,
    connect_influx_max_retries: str,
    connect_influx_retry_interval: str,
    connect_progress_enabled: str,
    timestamp: str,
    tags: str,
    remove_prefix: str,
) -> int:
    """Create an instance of the InfluxDB Sink connector.

    A list of topics can be specified using the TOPICLIST argument.
    If not, topics are discovered from Kafka. Use the ``--topic-regex`` and
    ``--excluded_topics`` options to help in selecting the topics
    that you want to write to InfluxDB. To check for new topics and update
    the connector configuration use the
    ``--auto-update`` and ``--check-interval`` options.
    """
    # Get configuration from the main command
    if ctx.parent:
        config = ctx.parent.obj["config"]
    # Connector configuration
    influx_config = InfluxConfig(
        name=name,
        connect_influx_url=connect_influx_url,
        connect_influx_db=connect_influx_db,
        tasks_max=int(tasks_max),
        connect_influx_username=connect_influx_username,
        connect_influx_password=connect_influx_password,
        connect_influx_error_policy=connect_influx_error_policy,
        connect_influx_max_retries=connect_influx_max_retries,
        connect_influx_retry_interval=connect_influx_retry_interval,
        connect_progress_enabled=(connect_progress_enabled == "true"),
        tags=tags,
        remove_prefix=remove_prefix,
    )
    # The variadic argument is a tuple
    topics: Set[str] = set(topiclist)
    if not topics:
        click.echo("Discoverying Kafka topics...")
        t = TopicNamesSet.from_kafka(config, topic_regex, excluded_topic_regex)
        topics = t.topic_names_set
        n = 0 if not topics else len(topics)
        click.echo(f"Found {n} topics.")
    connect = Connect(connect_url=config.connect_url)
    if topics:
        influx_config.update_config(topics, timestamp)
        # --validate option returns the validation results
        if validate:
            click.echo(
                connect.validate(
                    name=influx_config.connector_class,
                    connect_config=influx_config.asjson(),
                )
            )
            return 0
        # --dry-run option returns the connector configuration
        if dry_run:
            click.echo(influx_config.asjson())
            return 0
        # Validate configuration before creating the connector
        validation = connect.validate(
            name=influx_config.connector_class,
            connect_config=influx_config.asjson(),
        )
        try:
            error_count = json.loads(validation)["error_count"]
            click.echo(f"Validation returned {error_count} error(s).")
            if error_count > 0:
                click.echo(
                    "Use the ``--validate`` option to return the validation "
                    "results."
                )
                return 1
        except Exception:
            click.echo(validation)
            return 1
        click.echo(f"Uploading {name} connector configuration...")
        connect.create_or_update(
            name=name, connect_config=influx_config.asjson()
        )
    if auto_update:
        while True:
            time.sleep(int(check_interval) / 1000)
            try:
                # Current list of topics from Kafka
                t = TopicNamesSet.from_kafka(
                    config, topic_regex, excluded_topic_regex
                )
                current_topics = t.topic_names_set
                new_topics = list(set(current_topics) - set(topics))
                if new_topics:
                    click.echo("Found new topics, updating the connector...")
                    influx_config.update_config(current_topics, timestamp)
                    connect.create_or_update(
                        name=name, connect_config=influx_config.asjson()
                    )
                    topics = current_topics
            except KeyboardInterrupt:
                raise click.ClickException("Interruped.")
    return 0



"""The kafkaconnect package."""

__all__ = ["__version__", "metadata"]

import sys

if sys.version_info < (3, 8):
    from importlib_metadata import PackageNotFoundError, version
else:
    from importlib.metadata import PackageNotFoundError, version


__version__: str
"""The application version string of (PEP 440 / SemVer compatible)."""

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"
